using System;
using System.ComponentModel;
using System.Collections;
using System.Globalization;

#if NETCFDESIGNTIME
using System.ComponentModel;
using System.ComponentModel.Design.Serialization;
using System.Reflection;
#endif

namespace ReflectionIT {

	#if NETCFDESIGNTIME
	
	public class TimeConverter : TypeConverter {

		public override bool CanConvertFrom ( ITypeDescriptorContext context , System.Type sourceType ) {
			return (sourceType == typeof(string)) ;
		}

		public override object ConvertFrom(System.ComponentModel.ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value) {
			return new Time(Convert.ToString(value));
		}

		public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType) {
			if (destinationType.Equals(typeof(string)) || (destinationType == typeof(InstanceDescriptor))) {
				return true;
			} else {
				return base.CanConvertTo (context, destinationType);
			}
		}

		public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType) {
			if (destinationType == typeof(InstanceDescriptor)) {
				// This will make a Time a serialiable property in the IntializeComponent method
				ConstructorInfo ci = typeof(Time).GetConstructor(new Type[]{typeof(int), typeof(int)});
				Time t = (Time)value;
				return new InstanceDescriptor(ci,new object[]{t.Hours, t.Minutes});
			}

			if (destinationType.Equals(typeof(String))) {
				return value.ToString();
			} else {
				return base.ConvertTo (context, culture, value, destinationType);
			}
		}
	}
	#endif

	/// <summary>
	///    A Time comprises a number of Hours and a number of Minutes
	///    The following operators are defined:
	///    +(Time, Time)   adds two Times together
	///    +(Time, int)    adds a number of _minutes to a Time
	///    -(Time, Time)   subtracts one time from the other (the first must be bigger)
	///    -(Time, int)    subtracts a number of _minutes from a Time
	///    ++              advances the Time by one minute
	///    --              subtracts a minute from the Time
	///    ToString(Time)        returns a string representation of the Time
	///    Equals(Time, Time)    compares two Times, returns true if they are the same, false otherwise
	///    >, <, >=, <=, ==, !=  perform comparisons on Times
	///    int, float, Time      conversion operators
	/// </summary>
	#if NETCFDESIGNTIME
	[TypeConverter(typeof(TimeConverter))]
	[Serializable]
	#endif
	public struct Time : IComparable {
		// compiler generated no-arg struct constructor 
		// sets _hours and _minutes to zero

		private int _hours;
		private int _minutes;

		public Time(string time) : this(Convert.ToInt32(time.Substring(0, 2)), Convert.ToInt32(time.Substring(3, 2))) {
		}

		public Time(int minutes) : this(0, minutes) {
		}

		public Time(int hours, int minutes) {
			this._hours = hours;
			this._minutes = minutes;
			this.Normalize();
		}

		public Time(double hours) {
			this._hours = Convert.ToInt32(hours);
			this._minutes = Convert.ToInt32(60 * (hours - this._hours));
			this.Normalize();
		}

		public static Time Now {
			get { return new Time(DateTime.Now.Hour, DateTime.Now.Minute); }
		}

		// arithmetic
	
		public static Time operator+(Time lhs, Time rhs) {
			return new Time(lhs._hours + rhs._hours, 
				lhs._minutes + rhs._minutes);
		}

		public static Time operator+(Time t1, int _minutes) {
			return t1 + new Time(_minutes);
		}

		public static Time operator-(Time lhs, Time rhs) {
			return new Time(lhs.TotalMinutes() - rhs.TotalMinutes());
		}

		public static Time operator-(Time lhs, int _minutes) {
			return new Time(lhs.TotalMinutes() - _minutes);
		}

		public static Time operator++(Time t1) {
			return t1 + 1;
		}

		public static Time operator--(Time t1) {
			return t1 - 1;
		}

		// equality and relational

		public static bool operator==(Time lhs, Time rhs) {
			return lhs.Compare(rhs) == 0;
		}

		public static bool operator!=(Time lhs, Time rhs) {
			return lhs.Compare(rhs) != 0;
		}

		public static bool operator<(Time lhs, Time rhs) {
			return lhs.Compare(rhs) < 0;
		}

		public static bool operator>(Time lhs, Time rhs) {
			return lhs.Compare(rhs) > 0;
		}

		public static bool operator<=(Time lhs, Time rhs) {
			return lhs.Compare(rhs) <= 0;
		}

		public static bool operator>=(Time lhs, Time rhs) {
			return lhs.Compare(rhs) >= 0;
		}

		// conversions

		public static explicit operator Time (int _minutes) {
			return new Time(0, _minutes);
		}

		public static explicit operator Time (float _minutes) {
			return new Time(0, (int)_minutes);
		}

		public static implicit operator int (Time t1) {
			return t1.TotalMinutes();
		}

		public static explicit operator float (Time t1) {
			return t1.TotalMinutes();
		}

		public static implicit operator string (Time t1) {
			return t1.ToString();
		}

		// Inherited virtual methods (from Object)
	
		public override string ToString() {

			string s = CultureInfo.CurrentUICulture.DateTimeFormat.TimeSeparator;

			return String.Format("{0:00}{1}{2:00}", _hours, s, _minutes);
		}

		public override bool Equals(object obj) {
			return (obj is Time) && Compare((Time)obj) == 0;
		}

		public override int GetHashCode() {
			return TotalMinutes();
		}

		// Helper functions

		private int Compare(Time other) {
			long lhs = this.TotalMinutes();
			long rhs = other.TotalMinutes();
		
			return lhs.CompareTo(rhs);
		}
		
		int IComparable.CompareTo(object other) {
			return Compare((Time)other);
		}

		private void Normalize() {
			// This method normalizes the time.
			// For example if _hours was 10 and _minutes was 65, 
			// Normalize will change _hours to 11 and _minutes to 5
			if (_minutes < 0 || _hours < 0) {
				_minutes = 0;
				_hours = 0;
				throw new ArgumentException("Time too small");
			}
			_hours += (_minutes / 60);
			_minutes %= 60;
		}

		public int TotalMinutes() {
			return _hours * 60 + _minutes;
		}

		public Time AddHours(double value) {
			Time t = this;
			if (value < 0) {
				t = this - new Time(value * -1);
			} else {
				t = this + new Time(value);
			}
			return t;
		}

		public Time AddMinutes(int value) {
			Time t = this;
			if (value < 0) {
				t = this - new Time(value * -1);
			} else {
				t = this + new Time(value);
			}
			return t;
		}

		public int Hours {
			get { return _hours; }
		}

		public int Minutes {
			get { return _minutes; }
		}
	}
}
